from autolens.data.imaging import image as im
from autolens.data.array import grids
from autolens.lensing import ray_tracing
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.lensing.plotters import ray_tracing_plotters
from autolens.data.imaging.plotters import imaging_plotters

# In this example, we'll use the 'imaging' module (imported as 'im') to 'simulate' an regular of a strong lens made
# using a tracer_without_subhalo. By simulate, we mean that it will appear as if we had observed it using a real telescope,
# with this example making an regular representative of Hubble Space Telescope imaging.

# To simulate an regular, we need to model the telescope's optics. We'll do this by convolving the regular with a
# Point-Spread Function, which we can simulate as a Gaussian using the imaging module.
psf = im.PSF.simulate_as_gaussian(shape=(11, 11), pixel_scale=0.1, sigma=0.1)

# To simulate an regular, we use a special type of grid. This pads the grids 2D dimensions relative to the PSF-shape,
# to ensure that the edge's of our simulated regular are not degraded.
image_plane_grids = grids.DataGrids.grids_for_simulation(shape=(100, 100), pixel_scale=0.1, psf_shape=psf.shape)
print(image_plane_grids.regular.image_shape)
print(image_plane_grids.regular.padded_shape)

# Now, lets setup our lens model_galaxy, source model_galaxy and tracer_without_subhalo.
lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0))
source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.1, 0.1), axis_ratio=0.8, phi=45.0,
                                                        intensity=1.0, effective_radius=1.0, sersic_index=2.5))
tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                             image_plane_grids=[image_plane_grids])

# Lets look at the tracer_without_subhalo's regular-plane regular - this is the regular we'll be simulating.
ray_tracing_plotters.plot_image_plane_image(tracer=tracer)

# To simulate the regular, we don't use the regular-plane regular plotted above. Instead, we use an  regular-plane regular
# which has been generated specifically for simulating an regular, using the padded grid above that ensures
# edge-effects do not degrade our simulation's PSF convolution.
print(tracer.image_plane_image.shape)
print(tracer.image_plane_image_for_simulation.shape)

# Now, to simulate the regular, we pass the tracer_without_subhalo's regular-plane regular to the imaging module's simulate function. This
# adds the following effects:

# 1) Telescope optics: Using the Point Spread Function above.
# 2) The Background Sky: Although the regular that is returned is automatically background sky subtracted.
# 3) Poisson noise: Due to the background sky, lens model_galaxy and source model_galaxy Poisson photon counts.

image_simulated = im.Image.simulate(array=tracer.image_plane_image_for_simulation, pixel_scale=0.1,
                                               exposure_time=300.0, psf=psf, background_sky_level=0.1, add_noise=True)

# Lets plot the regular - we can see the regular has been blurred due to the telescope optics and noise has been added.
imaging_plotters.plot_image_subplot(image=image_simulated)

# Finally, lets output these files to.fits files, we'll begin to analyze them in the next tutorial!
# Unfortunately, in a Jupyter notebook you have to manually specify the path to PyAutoLens and this tutorial.
path = '/home/jammy/PyCharm/Projects/AutoLens/workspace/howtolens/chapter_1_introduction'
im.output_imaging_to_fits(image=image_simulated, image_path=path+'/data/regular.fits',
                                                 noise_map_path=path+'/data/noise_map.fits',
                                                 psf_path=path+'/data/psf.fits',
                          overwrite=True)

# You've just completed your third tutorial, try the following:

# 1) Change the size (sigma) of the PSF, what happens to the lensed source model_galaxy's appearance?
# 2) Increase the background sky level of the simulation - what happens to the regular?
# 3) Increase the expousre time of the simulation - what happens as you change this and the background sky?


import numpy as np

from autolens.model.profiles import light_and_mass_profiles as lmp, light_profiles as lp, mass_profiles as mp

grid = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [2.0, 4.0]])


class TestSersic(object):

    def test__grid_calculations__same_as_sersic(self):
        sersic_lp = lp.EllipticalSersic(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6,
                                        sersic_index=2.0)
        sersic_mp = mp.EllipticalSersic(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6,
                                        sersic_index=2.0, mass_to_light_ratio=2.0)
        sersic_lmp = lmp.EllipticalSersic(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6,
                                          sersic_index=2.0, mass_to_light_ratio=2.0)

        assert (sersic_lp.intensities_from_grid(grid) == sersic_lmp.intensities_from_grid(grid)).all()
        assert (sersic_mp.surface_density_from_grid(grid) == sersic_lmp.surface_density_from_grid(grid)).all()
        #    assert (sersic_mp.potential_from_grid(grid) == sersic_lmp.potential_from_grid(grid)).all()
        assert (sersic_mp.deflections_from_grid(grid) == sersic_lmp.deflections_from_grid(grid)).all()

    def test__spherical_and_elliptical_identical(self):
        elliptical = lmp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=1.0,
                                          effective_radius=1.0, sersic_index=2.0, mass_to_light_ratio=2.0)
        spherical = lmp.SphericalSersic(centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0, sersic_index=2.0,
                                        mass_to_light_ratio=2.0)

        assert (elliptical.intensities_from_grid(grid) == spherical.intensities_from_grid(grid)).all()
        assert (elliptical.surface_density_from_grid(grid) == spherical.surface_density_from_grid(grid)).all()
        # assert (elliptical.potential_from_grid(grid) == spherical.potential_from_grid(grid)).all()
        np.testing.assert_almost_equal(elliptical.deflections_from_grid(grid), spherical.deflections_from_grid(grid))


class TestExponential(object):

    def test__grid_calculations__same_as_exponential(self):
        sersic_lp = lp.EllipticalExponential(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6)
        sersic_mp = mp.EllipticalExponential(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6,
                                             mass_to_light_ratio=2.0)
        sersic_lmp = lmp.EllipticalExponential(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6,
                                               mass_to_light_ratio=2.0)

        assert (sersic_lp.intensities_from_grid(grid) == sersic_lmp.intensities_from_grid(grid)).all()
        assert (sersic_mp.surface_density_from_grid(grid) == sersic_lmp.surface_density_from_grid(grid)).all()
        #    assert (sersic_mp.potential_from_grid(grid) == sersic_lmp.potential_from_grid(grid)).all()
        assert (sersic_mp.deflections_from_grid(grid) == sersic_lmp.deflections_from_grid(grid)).all()

    def test__spherical_and_elliptical_identical(self):
        elliptical = lmp.EllipticalExponential(centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=1.0,
                                               effective_radius=1.0)
        spherical = lmp.SphericalExponential(centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0)

        assert (elliptical.intensities_from_grid(grid) == spherical.intensities_from_grid(grid)).all()
        assert (elliptical.surface_density_from_grid(grid) == spherical.surface_density_from_grid(grid)).all()
        # assert elliptical.potential_from_grid(grid) == spherical.potential_from_grid(grid)
        np.testing.assert_almost_equal(elliptical.deflections_from_grid(grid), spherical.deflections_from_grid(grid))


class TestDevVaucouleurs(object):

    def test__grid_calculations__same_as_dev_vaucouleurs(self):
        sersic_lp = lp.EllipticalDevVaucouleurs(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6)
        sersic_mp = mp.EllipticalDevVaucouleurs(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6,
                                                mass_to_light_ratio=2.0)
        sersic_lmp = lmp.EllipticalDevVaucouleurs(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6,
                                                  mass_to_light_ratio=2.0)

        assert (sersic_lp.intensities_from_grid(grid) == sersic_lmp.intensities_from_grid(grid)).all()
        assert (sersic_mp.surface_density_from_grid(grid) == sersic_lmp.surface_density_from_grid(grid)).all()
        #    assert (sersic_mp.potential_from_grid(grid) == sersic_lmp.potential_from_grid(grid)).all()
        assert (sersic_mp.deflections_from_grid(grid) == sersic_lmp.deflections_from_grid(grid)).all()

    def test__spherical_and_elliptical_identical(self):
        elliptical = lmp.EllipticalDevVaucouleurs(centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=1.0,
                                                  effective_radius=1.0)
        spherical = lmp.SphericalDevVaucouleurs(centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0)

        assert (elliptical.intensities_from_grid(grid) == spherical.intensities_from_grid(grid)).all()
        assert (elliptical.surface_density_from_grid(grid) == spherical.surface_density_from_grid(grid)).all()
        # assert elliptical.potential_from_grid(grid) == spherical.potential_from_grid(grid)
        np.testing.assert_almost_equal(elliptical.deflections_from_grid(grid), spherical.deflections_from_grid(grid))


class TestSersicRadialGradient(object):

    def test__grid_calculations__same_as_sersic_radial_gradient(self):
        sersic_lp = lp.EllipticalSersic(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6,
                                        sersic_index=2.0)
        sersic_mp = mp.EllipticalSersicRadialGradient(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6,
                                                      sersic_index=2.0, mass_to_light_ratio=2.0,
                                                      mass_to_light_gradient=0.5)
        sersic_lmp = lmp.EllipticalSersicRadialGradient(axis_ratio=0.7, phi=1.0, intensity=1.0, effective_radius=0.6,
                                                        sersic_index=2.0, mass_to_light_ratio=2.0,
                                                        mass_to_light_gradient=0.5)

        assert (sersic_lp.intensities_from_grid(grid) == sersic_lmp.intensities_from_grid(grid)).all()
        assert (sersic_mp.surface_density_from_grid(grid) == sersic_lmp.surface_density_from_grid(grid)).all()
        #    assert (sersic_mp.potential_from_grid(grid) == sersic_lmp.potential_from_grid(grid)).all()
        assert (sersic_mp.deflections_from_grid(grid) == sersic_lmp.deflections_from_grid(grid)).all()

    def test__spherical_and_elliptical_identical(self):
        elliptical = lmp.EllipticalSersicRadialGradient(centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=1.0,
                                                        effective_radius=1.0)
        spherical = lmp.SphericalSersicRadialGradient(centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0)

        assert (elliptical.intensities_from_grid(grid) == spherical.intensities_from_grid(grid)).all()
        assert (elliptical.surface_density_from_grid(grid) == spherical.surface_density_from_grid(grid)).all()
        # assert elliptical.potential_from_grid(grid) == spherical.potential_from_grid(grid)
        np.testing.assert_almost_equal(elliptical.deflections_from_grid(grid), spherical.deflections_from_grid(grid))



import matplotlib.pyplot as plt
from astropy.io import fits

from autolens import exc


def get_subplot_rows_columns_figsize(number_subplots):
    if number_subplots <= 2:
        return 1, 2, (18, 8)
    elif number_subplots <= 4:
        return 2, 2, (13, 10)
    elif number_subplots <= 6:
        return 2, 3, (18, 12)
    elif number_subplots <= 9:
        return 3, 3, (25, 20)
    elif number_subplots <= 12:
        return 3, 4, (25, 20)
    elif number_subplots <= 16:
        return 4, 4, (25, 20)
    elif number_subplots <= 20:
        return 4, 5, (25, 20)
    else:
        return 6, 6, (25, 20)


def setup_figure(figsize, as_subplot):
    if not as_subplot:
        plt.figure(figsize=figsize)


def set_title(title, titlesize):
    plt.title(title, fontsize=titlesize)


def output_figure(array, as_subplot, output_path, output_filename, output_format):
    if not as_subplot:

        if output_format is 'show':
            plt.show()
        elif output_format is 'png':
            plt.savefig(output_path + output_filename + '.png', bbox_inches='tight')
        elif output_format is 'fits':
            hdu = fits.PrimaryHDU()
            hdu.data = array
            hdu.writeto(output_path + output_filename + '.fits')


def output_subplot_array(output_path, output_filename, output_format):
    if output_format is 'show':
        plt.show()
    elif output_format is 'png':
        plt.savefig(output_path + output_filename + '.png', bbox_inches='tight')
    elif output_format is 'fits':
        raise exc.PlottingException('You cannot output a subplots with format .fits')


def close_figure(as_subplot):
    if not as_subplot:
        plt.close()



from matplotlib import pyplot as plt

from autofit import conf
from autolens.data.array.plotters import plotter_util
from autolens.data.imaging.plotters import imaging_plotters
from autolens.lensing.plotters import plane_plotters
from autolens.model.inversion.plotters import inversion_plotters
from autolens.data.fitting.plotters import fitting_plotters


def plot_fitting_subplot(fit, should_plot_mask=True, positions=None, should_plot_image_plane_pix=True,
                         units='arcsec', figsize=None, aspect='equal',
                         cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                         cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                         titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                         mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                         output_path=None, output_filename='lensing_fit', output_format='show', ignore_config=True):
    plot_lensing_fitting_as_subplot = conf.instance.general.get('output', 'plot_lensing_fitting_as_subplot', bool)

    if not plot_lensing_fitting_as_subplot and ignore_config is False:
        return

    if fit.tracer.total_planes == 1:

        if not fit.tracer.has_hyper_galaxy:

            plot_fitting_subplot_lens_plane_only(fit=fit, should_plot_mask=should_plot_mask, positions=positions,
                                                 should_plot_image_plane_pix=should_plot_image_plane_pix,
                                                 units=units, figsize=figsize,
                                                 aspect=aspect,
                                                 cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                                                 linthresh=linthresh,
                                                 linscale=linscale,
                                                 cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                                 titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                                 xyticksize=xyticksize,
                                                 mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                                                 grid_pointsize=grid_pointsize,
                                                 output_path=output_path, output_filename=output_filename,
                                                 output_format=output_format)
        elif fit.tracer.has_hyper_galaxy:

            plot_fitting_subplot_hyper_lens_plane_only(fit=fit, should_plot_mask=should_plot_mask, positions=positions,
                                                       should_plot_image_plane_pix=should_plot_image_plane_pix,
                                                       units=units, figsize=figsize,
                                                       aspect=aspect,
                                                       cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                                                       linthresh=linthresh,
                                                       linscale=linscale,
                                                       cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                                       titlesize=titlesize, xlabelsize=xlabelsize,
                                                       ylabelsize=ylabelsize,
                                                       xyticksize=xyticksize,
                                                       mask_pointsize=mask_pointsize,
                                                       position_pointsize=position_pointsize,
                                                       grid_pointsize=grid_pointsize,
                                                       output_path=output_path, output_filename=output_filename,
                                                       output_format=output_format)

    elif fit.tracer.total_planes == 2:

        if not fit.tracer.has_hyper_galaxy:
            plot_fitting_subplot_lens_and_source_planes(fit=fit, should_plot_mask=should_plot_mask, positions=positions,
                                                        should_plot_image_plane_pix=should_plot_image_plane_pix,
                                                        units=units, figsize=figsize,
                                                        aspect=aspect,
                                                        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                                                        linthresh=linthresh,
                                                        linscale=linscale,
                                                        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                                        titlesize=titlesize, xlabelsize=xlabelsize,
                                                        ylabelsize=ylabelsize,
                                                        xyticksize=xyticksize,
                                                        mask_pointsize=mask_pointsize,
                                                        position_pointsize=position_pointsize,
                                                        grid_pointsize=grid_pointsize,
                                                        output_path=output_path, output_filename=output_filename,
                                                        output_format=output_format)


def plot_fitting_subplot_lens_plane_only(fit, should_plot_mask=True, positions=None, should_plot_image_plane_pix=True,
                                         units='arcsec', figsize=None, aspect='equal',
                                         cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05,
                                         linscale=0.01,
                                         cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                                         titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                                         mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                                         output_path=None, output_filename='lensing_fit', output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lensing.fittingting.Fitter
        Class containing fitting between the model datas_ and observed lensing datas_ (including residuals, chi_squareds etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fittings)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fittings)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fittings) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=4)

    mask = fitting_plotters.get_mask(fit=fit, should_plot_mask=should_plot_mask)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    kpc_per_arcsec = fit.tracer.image_plane.kpc_per_arcsec_proper

    image_plane_pix_grid = get_image_plane_pix_grid(should_plot_image_plane_pix, fit)

    imaging_plotters.plot_image(image=fit.images[0], mask=mask, positions=positions,
                                image_plane_pix_grid=image_plane_pix_grid,
                                as_subplot=True,
                                units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                linscale=linscale,
                                cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                xyticksize=xyticksize,
                                grid_pointsize=grid_pointsize, position_pointsize=position_pointsize,
                                mask_pointsize=mask_pointsize,
                                output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 2)

    fitting_plotters.plot_model_image(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                      units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                      cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                      linscale=linscale,
                                      cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                      titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                      output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 3)

    fitting_plotters.plot_residuals(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                    units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                    cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                    linscale=linscale,
                                    cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                    titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                    output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 4)

    fitting_plotters.plot_chi_squareds(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                       units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                       cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                       linscale=linscale,
                                       cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                       output_path=output_path, output_filename='', output_format=output_format)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()


def plot_fitting_subplot_hyper_lens_plane_only(fit, should_plot_mask=True, positions=None,
                                               should_plot_image_plane_pix=True,
                                               units='arcsec', figsize=None, aspect='equal',
                                               cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05,
                                               linscale=0.01,
                                               cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                                               titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                                               mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                                               output_path=None, output_filename='lensing_hyper_fit', output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lensing.fittingting.Fitter
        Class containing fitting between the model datas_ and observed lensing datas_ (including residuals, chi_squareds etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fittings)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fittings)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fittings) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=9)

    mask = fitting_plotters.get_mask(fit=fit, should_plot_mask=should_plot_mask)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    kpc_per_arcsec = fit.tracer.image_plane.kpc_per_arcsec_proper

    image_plane_pix_grid = get_image_plane_pix_grid(should_plot_image_plane_pix, fit)

    imaging_plotters.plot_image(image=fit.images[0], mask=mask, positions=positions,
                                image_plane_pix_grid=image_plane_pix_grid, as_subplot=True,
                                units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                linscale=linscale,
                                cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                xyticksize=xyticksize,
                                grid_pointsize=grid_pointsize, position_pointsize=position_pointsize,
                                mask_pointsize=mask_pointsize,
                                output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 2)

    fitting_plotters.plot_model_image(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                      units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                      cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
                                      cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                      titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                      output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 3)

    fitting_plotters.plot_residuals(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                    units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                    cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
                                    cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                    titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                    output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 5)

    fitting_plotters.plot_chi_squareds(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                       units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                       cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                       linscale=linscale,
                                       cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                       output_path=output_path, output_filename='', output_format=output_format)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.subplot(rows, columns, 4)

    fitting_plotters.plot_contributions(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                        units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                        linscale=linscale,
                                        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                        output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 6)

    fitting_plotters.plot_scaled_chi_squareds(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                              linscale=linscale,
                                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                              titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                              output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 8)

    imaging_plotters.plot_noise_map(image=fit.images[0], mask=mask, as_subplot=True,
                                    units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                    cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                    linscale=linscale,
                                    cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                    titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                    xyticksize=xyticksize,
                                    output_path=output_path, output_format=output_format)

    plt.subplot(rows, columns, 9)

    fitting_plotters.plot_scaled_noise_map(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                           units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                           cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                           linscale=linscale,
                                           cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                           titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                           output_path=output_path, output_filename='', output_format=output_format)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()


def plot_fitting_subplot_lens_and_source_planes(fit, should_plot_mask=True, should_plot_source_grid=False, positions=None,
                                                should_plot_image_plane_pix=True,
                                                units='arcsec', figsize=None, aspect='equal',
                                                cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05,
                                                linscale=0.01,
                                                cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                                                titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                                                mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                                                output_path=None, output_filename='lensing_fit', output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lensing.fittingting.Fitter
        Class containing fitting between the model datas_ and observed lensing datas_ (including residuals, chi_squareds etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fittings)
    output_filename : str
        The name of the file that is output, if the output_type is a file format (e.g. png, fittings)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fittings) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=6)

    mask = fitting_plotters.get_mask(fit=fit, should_plot_mask=should_plot_mask)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    kpc_per_arcsec = fit.tracer.image_plane.kpc_per_arcsec_proper

    image_plane_pix_grid = get_image_plane_pix_grid(should_plot_image_plane_pix, fit)

    imaging_plotters.plot_image(image=fit.images[0], mask=mask, positions=positions,
                                image_plane_pix_grid=image_plane_pix_grid, as_subplot=True,
                                units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                linscale=linscale,
                                cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                xyticksize=xyticksize,
                                grid_pointsize=grid_pointsize, position_pointsize=position_pointsize,
                                mask_pointsize=mask_pointsize,
                                output_path=output_path, output_filename='', output_format=output_format)

    if fit.tracer.image_plane.has_light_profile:

        plt.subplot(rows, columns, 2)

        fitting_plotters.plot_model_image_of_plane(fit=fit, image_index=0, plane_index=0, mask=mask, as_subplot=True,
                                                   units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                                   cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                                   linscale=linscale,
                                                   cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                                   titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                                   xyticksize=xyticksize,
                                                   output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 3)

    fitting_plotters.plot_model_image_of_plane(fit=fit, image_index=0, plane_index=1, mask=mask, as_subplot=True,
                                               units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                               cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                               linscale=linscale,
                                               cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                               titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                               output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 4)

    if fit.total_inversions == 0:

        plane_plotters.plot_plane_image(plane=fit.tracer.source_plane, image_index=0, as_subplot=True,
                                        positions=None, plot_grid=should_plot_source_grid,
                                        units=units, figsize=figsize, aspect=aspect,
                                        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                        linscale=linscale,
                                        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                        xyticksize=xyticksize,
                                        grid_pointsize=grid_pointsize, position_pointsize=position_pointsize,
                                        output_path=output_path, output_filename='', output_format=output_format)

    elif fit.total_inversions == 1:

        inversion_plotters.plot_reconstructed_pixelization(inversion=fit.inversion, positions=None,
                                                           should_plot_grid=False, should_plot_centres=False,
                                                           as_subplot=True,
                                                           units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize,
                                                           aspect=aspect,
                                                           cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                                                           linthresh=linthresh, linscale=linscale,
                                                           cb_ticksize=cb_ticksize, cb_fraction=cb_fraction,
                                                           cb_pad=cb_pad,
                                                           titlesize=titlesize, xlabelsize=xlabelsize,
                                                           ylabelsize=ylabelsize, xyticksize=xyticksize,
                                                           output_path=output_path, output_filename=None,
                                                           output_format=output_format)

    plt.subplot(rows, columns, 5)

    fitting_plotters.plot_residuals(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                    units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                    cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
                                    cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                    titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                    output_path=output_path, output_filename='', output_format=output_format)

    plt.subplot(rows, columns, 6)

    fitting_plotters.plot_chi_squareds(fit=fit, image_index=0, mask=mask, as_subplot=True,
                                       units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                       cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                       linscale=linscale,
                                       cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                                       output_path=output_path, output_filename='', output_format=output_format)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()


def plot_fitting_individuals(fit, units='kpc', output_path=None, output_format='show'):
    if fit.tracer.total_planes == 1:

        if not fit.tracer.has_hyper_galaxy:

            plot_fitting_individuals_lens_plane_only(fit, units, output_path, output_format)


        elif fit.tracer.has_hyper_galaxy:

            plot_fitting_individuals_hyper_lens_plane_only(fit, units, output_path, output_format)

    elif fit.tracer.total_planes == 2:

        if not fit.tracer.has_hyper_galaxy:
            plot_fitting_individuals_lens_and_source_planes(fit, units, output_path, output_format)


def plot_fitting_individuals_lens_plane_only(fit, units='kpc', output_path=None, output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lensing.fittingting.Fitter
        Class containing fitting between the model datas_ and observed lensing datas_ (including residuals, chi_squareds etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fittings)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fittings) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    plot_lensing_fitting_model_image = conf.instance.general.get('output', 'plot_lensing_fitting_model_image', bool)
    plot_lensing_fitting_residuals = conf.instance.general.get('output', 'plot_lensing_fitting_residuals', bool)
    plot_lensing_fitting_chi_squareds = conf.instance.general.get('output', 'plot_lensing_fitting_chi_squareds', bool)

    mask = fitting_plotters.get_mask(fit=fit, should_plot_mask=True)

    kpc_per_arcsec = fit.tracer.image_plane.kpc_per_arcsec_proper

    if plot_lensing_fitting_model_image:
        fitting_plotters.plot_model_image(fit=fit, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                          output_path=output_path, output_format=output_format)

    if plot_lensing_fitting_residuals:
        fitting_plotters.plot_residuals(fit=fit, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                        output_path=output_path, output_format=output_format)

    if plot_lensing_fitting_chi_squareds:
        fitting_plotters.plot_chi_squareds(fit=fit, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                           output_path=output_path, output_format=output_format)


def plot_fitting_individuals_hyper_lens_plane_only(fit, units='kpc', output_path=None, output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lensing.fittingting.Fitter
        Class containing fitting between the model datas_ and observed lensing datas_ (including residuals, chi_squareds etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fittings)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fittings) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    plot_lensing_fitting_model_image = \
        conf.instance.general.get('output', 'plot_lensing_fitting_model_image', bool)
    plot_lensing_fitting_residuals = \
        conf.instance.general.get('output', 'plot_lensing_fitting_residuals', bool)
    plot_lensing_fitting_chi_squareds = \
        conf.instance.general.get('output', 'plot_lensing_fitting_chi_squareds', bool)
    plot_lensing_fitting_contributions = \
        conf.instance.general.get('output', 'plot_lensing_fitting_contributions', bool)
    plot_lensing_fitting_scaled_chi_squareds = \
        conf.instance.general.get('output', 'plot_lensing_fitting_scaled_chi_squareds', bool)
    plot_lensing_fitting_scaled_noise_map = \
        conf.instance.general.get('output', 'plot_lensing_fitting_scaled_noise_map', bool)

    mask = fitting_plotters.get_mask(fit=fit, should_plot_mask=True)

    kpc_per_arcsec = fit.tracer.image_plane.kpc_per_arcsec_proper

    if plot_lensing_fitting_model_image:
        fitting_plotters.plot_model_image(fit=fit, image_index=0, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                          output_path=output_path, output_format=output_format)

    if plot_lensing_fitting_residuals:
        fitting_plotters.plot_residuals(fit=fit, image_index=0, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                        output_path=output_path, output_format=output_format)

    if plot_lensing_fitting_chi_squareds:
        fitting_plotters.plot_chi_squareds(fit=fit, image_index=0, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                           output_path=output_path, output_format=output_format)

    if plot_lensing_fitting_contributions:
        fitting_plotters.plot_contributions(fit=fit, image_index=0, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                            output_path=output_path, output_format=output_format)

    if plot_lensing_fitting_scaled_noise_map:
        fitting_plotters.plot_scaled_noise_map(fit=fit, image_index=0, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                               output_path=output_path, output_format=output_format)

    if plot_lensing_fitting_scaled_chi_squareds:
        fitting_plotters.plot_scaled_chi_squareds(fit=fit, image_index=0, mask=mask, units=units,
                                                  kpc_per_arcsec=kpc_per_arcsec, output_path=output_path,
                                                  output_format=output_format)


def plot_fitting_individuals_lens_and_source_planes(fit, units='kpc', output_path=None, output_format='show'):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lensing.fittingting.Fitter
        Class containing fitting between the model datas_ and observed lensing datas_ (including residuals, chi_squareds etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fittings)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fittings) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    plot_lensing_fitting_model_image = \
        conf.instance.general.get('output', 'plot_lensing_fitting_model_image', bool)
    plot_lensing_fitting_lens_model_image = \
        conf.instance.general.get('output', 'plot_lensing_fitting_lens_model_image', bool)
    plot_lensing_fitting_source_model_image = \
        conf.instance.general.get('output', 'plot_lensing_fitting_source_model_image', bool)
    plot_lensing_fitting_source_plane_image = \
        conf.instance.general.get('output', 'plot_lensing_fitting_source_plane_image', bool)
    plot_lensing_fitting_residuals = conf.instance.general.get('output', 'plot_lensing_fitting_residuals', bool)
    plot_lensing_fitting_chi_squareds = conf.instance.general.get('output', 'plot_lensing_fitting_chi_squareds', bool)

    mask = fitting_plotters.get_mask(fit=fit, should_plot_mask=True)

    kpc_per_arcsec = fit.tracer.image_plane.kpc_per_arcsec_proper

    if plot_lensing_fitting_model_image:
        fitting_plotters.plot_model_image(fit=fit, image_index=0, units=units, mask=mask, kpc_per_arcsec=kpc_per_arcsec,
                                          output_path=output_path, output_format=output_format)

    if plot_lensing_fitting_lens_model_image:
        fitting_plotters.plot_model_image_of_plane(fit=fit, image_index=0, plane_index=0, mask=mask, units=units,
                                                   kpc_per_arcsec=kpc_per_arcsec,
                                                   output_path=output_path, output_filename='fit_lens_plane_model_image',
                                                   output_format=output_format)

    if plot_lensing_fitting_source_model_image:
        fitting_plotters.plot_model_image_of_plane(fit=fit, image_index=0, plane_index=1, mask=mask, units=units,
                                                   kpc_per_arcsec=kpc_per_arcsec,
                                                   output_path=output_path, output_filename='fit_source_plane_model_image',
                                                   output_format=output_format)

    if plot_lensing_fitting_source_plane_image:

        if fit.total_inversions == 0:

           plane_plotters.plot_plane_image(plane=fit.tracer.source_plane, image_index=0, plot_grid=True,
                                           units=units, figsize=(20, 20),
                                           output_path=output_path, output_filename='fit_source_plane',
                                           output_format=output_format)

        elif fit.total_inversions == 1:

            inversion_plotters.plot_reconstructed_pixelization(inversion=fit.inversion, should_plot_grid=True,
                                                               units=units, figsize=(20, 20),
                                                               output_path=output_path, output_filename='fit_source_plane',
                                                               output_format=output_format)

    if plot_lensing_fitting_residuals:
        fitting_plotters.plot_residuals(fit=fit, image_index=0, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                        output_path=output_path, output_format=output_format)

    if plot_lensing_fitting_chi_squareds:
        fitting_plotters.plot_chi_squareds(fit=fit, image_index=0, mask=mask, units=units, kpc_per_arcsec=kpc_per_arcsec,
                                           output_path=output_path, output_format=output_format)

def get_image_plane_pix_grid(should_plot_image_plane_pix, fit):
    if hasattr(fit, 'mapper'):
        if should_plot_image_plane_pix and fit.mapper.is_image_plane_pixelization:
            return fit.tracer.image_plane.grids[0].pix
    else:
        return None


import numpy as np
import pytest

from autolens.data.imaging import image as im
from autolens.data.array import mask as msk
from autolens.data.fitting import fitting_data as fit_data, fitting
from autolens.model.galaxy import galaxy as g
from autolens.lensing import lensing_image
from test.mock.mock_galaxy import MockHyperGalaxy

@pytest.fixture(name='fi_no_blur')
def make_li_no_blur():

    image = np.array([[0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0]])

    psf = im.PSF(array=(np.array([[0.0, 0.0, 0.0],
                                     [0.0, 1.0, 0.0],
                                     [0.0, 0.0, 0.0]])), pixel_scale=1.0, renormalize=False)
    image = im.Image(image, pixel_scale=1.0, psf=psf, noise_map=np.ones((4, 4)))

    ma = np.array([[True, True, True, True],
                   [True, False, False, True],
                   [True, False, False, True],
                   [True, True, True, True]])
    ma = msk.Mask(array=ma, pixel_scale=1.0)

    return lensing_image.LensingImage(image, ma, sub_grid_size=2)


@pytest.fixture(name='fi_blur')
def make_li_blur():

    image = np.array([[0.0, 0.0, 0.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 1.0, 1.0, 0.0],
                   [0.0, 0.0, 0.0, 0.0]])

    psf = im.PSF(array=(np.array([[1.0, 1.0, 1.0],
                                     [1.0, 1.0, 1.0],
                                     [1.0, 1.0, 1.0]])), pixel_scale=1.0, renormalize=False)
    image = im.Image(image, pixel_scale=1.0, psf=psf, noise_map=np.ones((4, 4)))

    ma = np.array([[True, True, True, True],
                   [True, False, False, True],
                   [True, False, False, True],
                   [True, True, True, True]])
    ma = msk.Mask(array=ma, pixel_scale=1.0)

    return lensing_image.LensingImage(image, ma, sub_grid_size=2)





import os
import shutil

import pytest
import numpy as np

from autofit import conf
from autolens.data.array import grids, mask as msk
from autolens.lensing.plotters import plane_plotters
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from autolens.model.galaxy import galaxy as g
from autolens.lensing import plane as pl


@pytest.fixture(name='general_config')
def test_general_config():
    general_config_path = "{}/../../test_files/configs/plotting/".format(os.path.dirname(os.path.realpath(__file__)))
    conf.instance.general = conf.NamedConfig(general_config_path + "general.ini")


@pytest.fixture(name='positions')
def test_positions():
    positions = [[[0.1, 0.1], [0.2, 0.2]], [[0.3, 0.3]]]
    return list(map(lambda position_set: np.asarray(position_set), positions))

@pytest.fixture(name='mask')
def test_mask():
    return msk.Mask.circular(shape=((3,3)), pixel_scale=0.1, radius_arcsec=0.1)

@pytest.fixture(name='plane_plotter_path')
def test_plane_plotter_setup():
    plane_plotter_path = "{}/../../test_files/plotting/plane/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(plane_plotter_path):
        shutil.rmtree(plane_plotter_path)

    os.mkdir(plane_plotter_path)

    return plane_plotter_path


@pytest.fixture(name='galaxy_light')
def test_galaxy_light():
    return g.Galaxy(light=lp.EllipticalSersic(intensity=1.0))


@pytest.fixture(name='galaxy_mass')
def test_galaxy_mass():
    return g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0))


@pytest.fixture(name='grids')
def test_grids():
    return grids.DataGrids.from_shape_and_pixel_scale(shape=(100, 100), pixel_scale=0.05, sub_grid_size=2)

@pytest.fixture(name='plane')
def test_plane(galaxy_light, grids):
    return pl.Plane(galaxies=[galaxy_light], grids=[grids])



def test__image_plane_image_is_output(plane, mask, positions, plane_plotter_path):
    plane_plotters.plot_image_plane_image(plane=plane, mask=mask, positions=positions,
                                          output_path=plane_plotter_path, output_format='png')
    assert os.path.isfile(path=plane_plotter_path + 'plane_image_plane_image.png')
    os.remove(path=plane_plotter_path + 'plane_image_plane_image.png')

def test__plane_image_is_output(plane, positions, plane_plotter_path):
    plane_plotters.plot_plane_image(plane=plane, positions=positions, output_path=plane_plotter_path,
                                    output_format='png')
    assert os.path.isfile(path=plane_plotter_path + 'plane_image.png')
    os.remove(path=plane_plotter_path + 'plane_image.png')

def test__surface_density_is_output(plane, plane_plotter_path):
    
    plane_plotters.plot_surface_density(plane=plane, output_path=plane_plotter_path, output_format='png')
    assert os.path.isfile(path=plane_plotter_path + 'plane_surface_density.png')
    os.remove(path=plane_plotter_path + 'plane_surface_density.png')

def test__potential_is_output(plane, plane_plotter_path):
    
    plane_plotters.plot_potential(plane=plane, output_path=plane_plotter_path, output_format='png')
    assert os.path.isfile(path=plane_plotter_path + 'plane_potential.png')
    os.remove(path=plane_plotter_path + 'plane_potential.png')

def test__deflections_y_is_output(plane, plane_plotter_path):
    
    plane_plotters.plot_deflections_y(plane=plane, output_path=plane_plotter_path, output_format='png')
    assert os.path.isfile(path=plane_plotter_path + 'plane_deflections_y.png')
    os.remove(path=plane_plotter_path + 'plane_deflections_y.png')

def test__deflections_x_is_output(plane, plane_plotter_path):
    
    plane_plotters.plot_deflections_x(plane=plane, output_path=plane_plotter_path, output_format='png')
    assert os.path.isfile(path=plane_plotter_path + 'plane_deflections_x.png')
    os.remove(path=plane_plotter_path + 'plane_deflections_x.png')

def test__plane_grid_is_output(plane, plane_plotter_path):

    plane_plotters.plot_plane_grid(plane=plane, output_path=plane_plotter_path, output_format='png')
    assert os.path.isfile(path=plane_plotter_path + 'plane_grid.png')
    os.remove(path=plane_plotter_path + 'plane_grid.png')


import numpy as np
import scipy.spatial
import sklearn.cluster

from autolens import exc
from autolens.data.array import grids, scaled_array
from autolens.model.inversion import mappers
from autolens.model.inversion.util import pixelization_util


def setup_image_plane_pixelization_grid_from_galaxies_and_grids(galaxies, grids):
    """An image-plane pixelization is one where its pixel centres are computed by tracing a sparse grid of pixels from \
    the data's regular grid to other planes (e.g. the source-plane).

    Provided a galaxy has an image-plane pixelization, this function returns a new *DataGrids* instance where the \
    image-plane pixelization's sparse grid is added to it as an attibute.

    Thus, when the *DataGrids* are are passed to the *ray_tracing* module this sparse grid is also traced and the \
    traced coordinates represent the centre of each pixelization pixel.

    Parameters
    -----------
    galaxies : [model.galaxy.galaxy.Galaxy]
        A list of galaxies, which may contain pixelizations and an *ImagePlanePixelization*.
    grids : data.array.grids.DataGrids
        The collection of grids (regular, sub, etc.) which the image-plane pixelization grid (referred to as pix) \
        may be added to.
    """
    for galaxy in galaxies:
        if hasattr(galaxy, 'pixelization'):
            if isinstance(galaxy.pixelization, ImagePlanePixelization):

                image_plane_pix_grid = galaxy.pixelization.image_plane_pix_grid_from_regular_grid(
                    regular_grid=grids.regular)
                return grids.data_grids_with_pix_grid(pix_grid=image_plane_pix_grid.sparse_grid,
                       regular_to_nearest_regular_pix=image_plane_pix_grid.regular_to_sparse)

    return grids


class ImagePlanePixelization(object):

    def __init__(self, shape):
        """An image-plane pixelization is one where its pixel centres are computed by tracing a sparse grid of pixels \
        from the data's regular grid to other planes (e.g. the source-plane).

        The traced coordinates of this sparse grid represent each centre of a pixelization pixel.

        See *grids.SparseToRegularGrid* for details on how this grid is calculated.

        Parameters
        -----------
        shape : (float, float) or (int, int)
            The shape of the image-plane pixelizaton grid in pixels (floats are converted to integers). The grid is \
            laid over the masked data such that it spans the most outer pixels of the mask.
        """
        self.shape = (int(shape[0]), int(shape[1]))

    def image_plane_pix_grid_from_regular_grid(self, regular_grid):
        """Calculate the image-plane pixelization from a regular-grid of coordinates and its masked.

        See *grids.SparseToRegularGrid* for details on how this grid is calculated.

        Parameters
        -----------
        regular_grid : grids.RegularGrid
            The grid of (y,x) arc-second coordinates at the centre of every data value (e.g. image-pixels).
        """
        image_pixel_scale = regular_grid.mask.pixel_scale
        pixel_scales = ((regular_grid.masked_shape_arcsec[0] + image_pixel_scale) / self.shape[0],
                        (regular_grid.masked_shape_arcsec[1] + image_pixel_scale) / self.shape[1])
        return grids.SparseToRegularGrid(unmasked_sparse_grid_shape=self.shape, pixel_scales=pixel_scales,
                                         regular_grid=regular_grid)


class Pixelization(object):

    def __init__(self):
        """ Abstract base class for a pixelization, which discretizes a set of coordinates (e.g. an datas_-grid) into \
        pixels.
        """

    def mapper_from_grids_and_border(self, grids, border):
        raise NotImplementedError("pixelization_mapper_from_grids_and_borders should be overridden")

    def __str__(self):
        return "\n".join(["{}: {}".format(k, v) for k, v in self.__dict__.items()])

    def __repr__(self):
        return "{}\n{}".format(self.__class__.__name__, str(self))


class Rectangular(Pixelization):

    def __init__(self, shape=(3, 3)):
        """A rectangular pixelization, where pixels are defined on a Cartesian and uniform grid of shape (rows, columns).

        Like an datas_, the indexing of the rectangular grid begins in the top-left corner and goes right and down.

        Parameters
        -----------
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
        """

        if shape[0] <= 2 or shape[1] <= 2:
            raise exc.PixelizationException('The rectangular pixelization must be at least dimensions 3x3')

        self.shape = (int(shape[0]), int(shape[1]))
        self.pixels = self.shape[0] * self.shape[1]
        super(Rectangular, self).__init__()

    class Geometry(scaled_array.RectangularArrayGeometry):

        def __init__(self, shape, pixel_scales, origin, pixel_neighbors, pixel_neighbors_size):
            """The geometry of a rectangular grid

            Parameters
            -----------
            shape : (int, int)
                The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
            pixel_scales : (float, float)
                The pixel-to-arcsecond scale of a pixel in the y and x directions.
            """
            self.shape = shape
            self.pixel_scales = pixel_scales
            self.origin = origin
            self.pixel_neighbors = pixel_neighbors.astype('int')
            self.pixel_neighbors_size = pixel_neighbors_size.astype('int')

        @property
        def pixel_centres(self):
            return self.grid_1d

    def geometry_from_grid(self, grid, buffer=1e-8):
        """Determine the geometry of the rectangular grid, by alligning it with the outer-most pixels on a grid \
        plus a small buffer.

        Parameters
        -----------
        grid : [[float, float]]
            The x and y pix grid (or sub-coordinates) which are to be matched with their pixels.
        buffer : float
            The size the grid-geometry is extended beyond the most exterior grid.
        """
        y_min = np.min(grid[:, 0]) - buffer
        y_max = np.max(grid[:, 0]) + buffer
        x_min = np.min(grid[:, 1]) - buffer
        x_max = np.max(grid[:, 1]) + buffer
        pixel_scales = (float((y_max - y_min) / self.shape[0]), float((x_max - x_min) / self.shape[1]))
        origin = ((y_max + y_min) / 2.0, (x_max + x_min) / 2.0)
        pixel_neighbors, pixel_neighbors_size = self.neighbors_from_pixelization()
        return self.Geometry(shape=self.shape, pixel_scales=pixel_scales, origin=origin,
                             pixel_neighbors=pixel_neighbors, pixel_neighbors_size=pixel_neighbors_size)

    def neighbors_from_pixelization(self):
        return pixelization_util.rectangular_neighbors_from_shape(shape=self.shape)

    def mapper_from_grids_and_border(self, grids, border):
        """Setup the pixelization mapper of this rectangular pixelization as follows:

        This first relocateds all grid-coordinates, such that any which tracer_normal beyond its border (e.g. due to high \
        levels of demagnification) are relocated to the border.

        Parameters
        ----------
        grids: mask.DataGrids
            A collection of grid describing the observed datas_'s pixel coordinates (includes an datas_ and sub grid).
        border : mask.ImagingGridBorders
            The border of the grids (defined by their datas_-plane masks).
        """

        if border is not None:
            relocated_grids = border.relocated_grids_from_grids(grids)
        else:
            relocated_grids = grids

        geometry = self.geometry_from_grid(grid=relocated_grids.sub)

        return mappers.RectangularMapper(pixels=self.pixels, grids=relocated_grids, border=border,
                                         shape=self.shape, geometry=geometry)


class Voronoi(Pixelization):

    def __init__(self):
        """Abstract base class for a Voronoi pixelization, which represents pixels as an irregular grid of Voronoi \
         pixels which can form any shape, size or tesselation.

         The traced datas_-pixels are paired to Voronoi pixels as the nearest-neighbors of the Voronoi pixel-centers.

         Parameters
         ----------
         pixels : int
             The number of pixels in the pixelization.
         """
        super(Voronoi, self).__init__()

    class Geometry(scaled_array.ArrayGeometry):

        def __init__(self, shape_arc_seconds, pixel_centres, origin, pixel_neighbors, pixel_neighbors_size):
            """The geometry of a rectangular grid

            Parameters
            -----------
            shape : (int, int)
                The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
            pixel_scales : (float, float)
                The pixel-to-arcsecond scale of a pixel in the y and x directions.
            """
            self.shape_arc_sec = shape_arc_seconds
            self.pixel_centres = pixel_centres
            self.origin = origin
            self.pixel_neighbors = pixel_neighbors.astype('int')
            self.pixel_neighbors_size = pixel_neighbors_size.astype('int')


    def geometry_from_grid(self, grid, pixel_centres, pixel_neighbors, pixel_neighbors_size, buffer=1e-8):
        """Determine the geometry of the rectangular grid, by alligning it with the outer-most pixels on a grid \
        plus a small buffer.

        Parameters
        -----------
        grid : [[float, float]]
            The x and y pix grid (or sub-coordinates) which are to be matched with their pixels.
        pixel_neighbors : [[]]
            The neighboring pix_pixels of each pix_pixel, computed via the Voronoi grid_coords. \
            (e.g. if the fifth pix_pixel neighbors pix_pixels 7, 9 and 44, pixel_neighbors[4] = [6, 8, 43])
        buffer : float
            The size the grid-geometry is extended beyond the most exterior grid.
        """
        y_min = np.min(grid[:, 0]) - buffer
        y_max = np.max(grid[:, 0]) + buffer
        x_min = np.min(grid[:, 1]) - buffer
        x_max = np.max(grid[:, 1]) + buffer
        shape_arc_seconds = (y_max - y_min, x_max - x_min)
        origin = ((y_max + y_min) / 2.0, (x_max + x_min) / 2.0)
        return self.Geometry(shape_arc_seconds=shape_arc_seconds, pixel_centres=pixel_centres, origin=origin,
                             pixel_neighbors=pixel_neighbors, pixel_neighbors_size=pixel_neighbors_size)

    @staticmethod
    def voronoi_from_pixel_centers(pixel_centers):
        """Compute the Voronoi grid of the pixelization, using the pixel centers.

        Parameters
        ----------
        pixel_centers : ndarray
            The x and y regular_grid to derive the Voronoi grid_coords.
        """
        return scipy.spatial.Voronoi(pixel_centers, qhull_options='Qbb Qc Qx Qm')


    def neighbors_from_pixelization(self, pixels, ridge_points):
        """Compute the neighbors of every pixel as a list of the pixel index's each pixel shares a vertex with.

        The ridge points of the Voronoi grid are used to derive this.

        Parameters
        ----------
        ridge_points : scipy.spatial.Voronoi.ridge_points
            Each Voronoi-ridge (two indexes representing a pixel mapping_matrix).
        """
        return pixelization_util.voronoi_neighbors_from_pixels_and_ridge_points(pixels=pixels,
                                                                                ridge_points=np.asarray(ridge_points))


class AdaptiveMagnification(Voronoi, ImagePlanePixelization):

    def __init__(self, shape=(3, 3)):
        """A Voronoi pixelization, which traces an image-plane grid to determine the cluster-centers.

        Parameters
        ----------
        shape : (int, int)
            The shape of the regular-grid whose centres form the centres of pixelization pixels.
        """
        super(AdaptiveMagnification, self).__init__()
        ImagePlanePixelization.__init__(self=self, shape=shape)

    def mapper_from_grids_and_border(self, grids, border):
        """Setup the pixelization mapper of the cluster pixelization.

        This first relocateds all grid-coordinates, such that any which tracer_normal beyond its border (e.g. due to high \
        levels of demagnification) are relocated to the border.

        Parameters
        ----------
        grids: mask.DataGrids
            A collection of grid describing the observed datas_'s pixel coordinates (includes an datas_ and sub grid).
        border : mask.ImagingGridBorders
            The border of the grids (defined by their datas_-plane masks).
        pixel_centres : ndarray
            The center of each Voronoi pixel, computed from an traced datas_-plane grid.
        image_to_nearest_image_pix : ndarray
            The mapping of each datas_ pixel to Voronoi pixels.
        """

        if border is not None:
            relocated_grids = border.relocated_grids_from_grids(grids)
        else:
            relocated_grids = grids

        pixel_centres = relocated_grids.pix
        pixels = pixel_centres.shape[0]

        voronoi = self.voronoi_from_pixel_centers(pixel_centres)

        pixel_neighbors, pixel_neighbors_size = self.neighbors_from_pixelization(pixels=pixels,
                                                                                 ridge_points=voronoi.ridge_points)
        geometry = self.geometry_from_grid(grid=relocated_grids.sub, pixel_centres=pixel_centres,
                                           pixel_neighbors=pixel_neighbors,
                                           pixel_neighbors_size=pixel_neighbors_size)

        return mappers.VoronoiMapper(pixels=pixels, grids=relocated_grids, border=border,
                                     voronoi=voronoi, geometry=geometry)


class Amorphous(Voronoi):

    def __init__(self, pix_grid_shape):
        """
        An amorphous pixelization, which represents pixels as a set of centers where all of the \
        nearest-neighbor pix-grid (i.e. traced masked_image-pixels) are mapped to them.

        For this pixelization, a set of cluster-pixels (defined in the masked_image-plane as a cluster uniform grid of \
        masked_image-pixels) are used to determine a set of pix-plane grid. These grid are then fed into a \
        weighted k-means clustering algorithm, such that the pixel centers adapt to the unlensed pix \
        surface-brightness profile.

        Parameters
        ----------
        pix_grid_shape : (int, int)
            The shape of the regular-grid whose centres form the centres of pixelization pixels.
        """
        super(Amorphous, self).__init__()

    def kmeans_cluster(self, pixels, cluster_grid):
        """Perform k-means clustering on the cluster_grid to compute the k-means clusters which represent \
        pixels.

        Parameters
        ----------
        cluster_grid : ndarray
            The x and y cluster-grid which are used to derive the k-means pixelization.
        """
        kmeans = sklearn.cluster.KMeans(pixels)
        km = kmeans.fit(cluster_grid)
        return km.cluster_centers_, km.labels_


from autolens.data.fitting import fitting
from autolens.data.fitting.util import fitting_util
from autolens.lensing import ray_tracing

class AbstractSensitivityFit(object):

    def __init__(self, tracer_normal, tracer_sensitive):

        self.tracer_normal = tracer_normal
        self.tracer_sensitive = tracer_sensitive


class SensitivityProfileFit(AbstractSensitivityFit):

    def __init__(self, sensitivity_images, tracer_normal, tracer_sensitive):
        """
        Class to evaluate the fit between a model described by a tracer_normal and an actual sensitivity_image.

        Parameters
        ----------
        sensitivity_images: [li.LensingImage]
            An sensitivity_image that has been masked for efficiency
        tracer_normal: ray_tracing.AbstractTracer
            An object describing the model
        """
        AbstractSensitivityFit.__init__(self=self, tracer_normal=tracer_normal, tracer_sensitive=tracer_sensitive)

        self.fit_normal = fitting.AbstractConvolutionFit(fitting_images=sensitivity_images,
                                                         images_=tracer_normal.image_plane_images_,
                                                         blurring_images_=tracer_normal.image_plane_blurring_images_)
        
        self.fit_sensitive = fitting.AbstractConvolutionFit(fitting_images=sensitivity_images,
                                                            images_=tracer_sensitive.image_plane_images_,
                                                            blurring_images_=tracer_sensitive.image_plane_blurring_images_)

    @classmethod
    def fast_likelihood(cls, sensitivity_images, tracer_normal, tracer_sensitive):
        """
        Fast calculation of likelihood

        Parameters
        ----------
        sensitivity_images: [li.LensingImage]
            An sensitivity_image that has been masked for efficiency
        tracer_normal: ray_tracing.AbstractTracer
            An object describing the model
        """

        convolvers = list(map(lambda lensing_image : lensing_image.convolver_image, sensitivity_images))
        noise_maps_ = list(map(lambda lensing_image : lensing_image.noise_map_, sensitivity_images))
        
        model_images_normal_ = fitting_util.blur_images_including_blurring_regions(images_=tracer_normal.image_plane_images_,
                                                                              blurring_images_=tracer_normal.image_plane_blurring_images_, convolvers=convolvers)
        
        residuals_normal_ = fitting_util.residuals_from_datas_and_model_datas(datas_=sensitivity_images,
                                                                         model_datas_=model_images_normal_)
        
        chi_squareds_normal_ = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals_=residuals_normal_,
                                                                                  noise_maps_=noise_maps_)
        
        chi_squared_terms_normal = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=chi_squareds_normal_)
        
        noise_terms_normal = fitting_util.noise_terms_from_noise_maps(noise_maps_=noise_maps_)
        likelihoods_normal = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms=chi_squared_terms_normal,
                                                                                   noise_terms=noise_terms_normal)
        
        model_images_sensitive_ = fitting_util.blur_images_including_blurring_regions(
            images_=tracer_sensitive.image_plane_images_, blurring_images_=tracer_sensitive.image_plane_blurring_images_,
            convolvers=convolvers)
        
        residuals_sensitive_ = fitting_util.residuals_from_datas_and_model_datas(datas_=sensitivity_images,
                                                                            model_datas_=model_images_sensitive_),
        chi_squareds_sensitive_ = fitting_util.chi_squareds_from_residuals_and_noise_maps(residuals_=residuals_sensitive_,
                                                                                     noise_maps_=noise_maps_),
        chi_squared_terms_sensitive = fitting_util.chi_squared_terms_from_chi_squareds(chi_squareds_=chi_squareds_sensitive_)
        noise_terms_sensitive = fitting_util.noise_terms_from_noise_maps(noise_maps_=noise_maps_)
        
        likelihoods_sensitive = fitting_util.likelihoods_from_chi_squareds_and_noise_terms(chi_squared_terms=chi_squared_terms_sensitive,
                                                                                      noise_terms=noise_terms_sensitive)

        return sum(likelihoods_sensitive) - sum(likelihoods_normal)

    @property
    def likelihood(self):
        return self.fit_sensitive.likelihood - self.fit_normal.likelihood


from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy as g
from autolens.lensing import ray_tracing
from autolens.data.array import grids
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion.plotters import mapper_plotters

# We'll start by learning about pixelizations, which we typically apply to a source-plane (but could, if we wanted,
# apply to an regular-plane).
#
# Lets setup a lensed source-plane grid, using a lens model_galaxy and tracer_normal.
image_plane_grids = grids.DataGrids.from_shape_and_pixel_scale(shape=(100, 100), pixel_scale=0.05,
                                                               sub_grid_size=2)
lens_galaxy = g.Galaxy(mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, einstein_radius=1.6))

# (Our source model_galaxy doesn't have a light profile from here on, as we're reconstructing its light using a pixelization).
tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[g.Galaxy()],
                                             image_plane_grids=[image_plane_grids])

# Next, lets set up a pixelization using the 'pixelizations' module, which we've imported as 'pix'.
# There are multiple pixelizations available in PyAutoLens, but for now we'll keep it simple and use a uniform
# rectangular grid. As usual, the grid's 'shape' defines its (y,x) dimensions.
rectangular = pix.Rectangular(shape=(25, 25))

# By itself, a pixelization doesn't tell us much. It has no grid of coordinates, no regular, and nothing which tells it
# about the lens we're fitting. This information comes when we use the pixelization to set up a 'mapper'.
# (The 'border=None' will be covered in tutorial 5, so just ignore it for now!)
mapper = rectangular.mapper_from_grids_and_border(grids=tracer.source_plane.grids[0], border=None)

# This mapper is a 'RectangularMapper' - every pixelization generates it owns mapper.
print(type(mapper))

# By plotting our mapper, we now see our pixelization. Its a fairly boring grid of rectangular pixels.
mapper_plotters.plot_rectangular_mapper(mapper=mapper, should_plot_grid=False,
                                        title='Fairly Boring Grid of Rectangular Pixels')

# However, the mapper does contain lots of interesting information about our pixelization, for example its geometry
# attribute tells us where the pixel centers are located
print('Rectangular Grid Pixel Centre 1:')
print(mapper.geometry.pixel_centres[0])
print('Rectangular Grid Pixel Centre 2:')
print(mapper.geometry.pixel_centres[1])
print('Rectangular Grid Pixel Centre 3:')
print(mapper.geometry.pixel_centres[2])
print('etc.')

# Infact, we can plot these origin on our grid - to make it look slightly less boring!
mapper_plotters.plot_rectangular_mapper(mapper=mapper, should_plot_grid=False, should_plot_centres=True)

# The mapper also has the (source-plane) grid that we passed when we set it up. Lets check they're the same grids.
print('Source Grid Pixel 1')
print(tracer.source_plane.grids[0].regular[0])
print(mapper.grids.regular[0])
print('Source Grid Pixel 2')
print(tracer.source_plane.grids[0].regular[1])
print(mapper.grids.regular[1])
print('etc.')

# We can over-lay the grid on top. Its starting too look a bit less boring now!
mapper_plotters.plot_rectangular_mapper(mapper=mapper, should_plot_centres=True, should_plot_grid=True)

# Finally, the mapper and its geometry has lots more information about the pixelization, for example, the arc-second
# size and dimensions.
print(mapper.geometry.shape_arc_seconds)
print(mapper.geometry.arc_second_maxima)
print(mapper.geometry.arc_second_minima)

# And with that, we're done. This was a relatively gentle introduction in the world of pixelizations, but one that
# was hopefully easy to follow. Think about the following questions before moving on to the next tutorial:

# 1) Look at how the source-grid coordinates are distributed over the rectangular pixel-grid. Are these points
#    distributed evenly over the rectangular grid's pixels? Do some pixels have a lot more grid-points inside of them?
#    Do some pixels have no grid-points in them?

#  2) The rectangular pixelization's edges are aligned with the most exterior coordinates of the source-grid. This is
#     intentional - why do you think this is?


from matplotlib import pyplot as plt

from autofit import conf
from autolens.data.array.plotters import plotter_util, array_plotters


def plot_image_subplot(image, plot_origin=True, mask=None, should_plot_border=False, positions=None,
                       image_plane_pix_grid=None,
                       units='arcsec', kpc_per_arcsec=None, figsize=None, aspect='equal',
                       cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                       cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                       titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                       mask_pointsize=10, position_pointsize=30, grid_pointsize=1,
                       output_path=None, output_filename='regular', output_format='show', ignore_config=True):
    """Plot the observed datas_ of an analysis, using the *Image* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    figsize
    ignore_config
    output_filename
    units
    positions
    image : autolens.imaging.regular.Image
        Class containing the datas_, noise-mappers and PSF that are to be plotted.
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    plot_image_as_subplot = conf.instance.general.get('output', 'plot_imaging_as_subplot', bool)

    if plot_image_as_subplot or ignore_config:

        rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=4)

        if figsize is None:
            figsize = figsize_tool

        plt.figure(figsize=figsize)
        plt.subplot(rows, columns, 1)

        plot_image(image=image, plot_origin=plot_origin, mask=mask, should_plot_border=should_plot_border,
                   positions=positions, image_plane_pix_grid=image_plane_pix_grid, as_subplot=True,
                   units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                   cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
                   cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                   titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                   mask_pointsize=mask_pointsize, position_pointsize=position_pointsize, grid_pointsize=grid_pointsize,
                   output_path=output_path, output_format=output_format)

        plt.subplot(rows, columns, 2)

        plot_noise_map(image=image, plot_origin=plot_origin, mask=mask, as_subplot=True,
                       units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                       cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                       linscale=linscale,
                       cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                       mask_pointsize=mask_pointsize,
                       output_path=output_path, output_format=output_format)

        plt.subplot(rows, columns, 3)

        plot_psf(image=image, as_subplot=True,
                 units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                 cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
                 cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                 titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                 output_path=output_path, output_format=output_format)

        plt.subplot(rows, columns, 4)

        plot_signal_to_noise_map(image=image, plot_origin=plot_origin, mask=mask, as_subplot=True,
                                 units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                                 cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                 linscale=linscale,
                                 cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                 titlesize=titlesize, xlabelsize=xlabelsize,
                                 ylabelsize=ylabelsize, xyticksize=xyticksize,
                                 mask_pointsize=mask_pointsize,
                                 output_path=output_path, output_format=output_format)

        plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                          output_format=output_format)

        plt.close()


def plot_image_individual(image, plot_origin=True, mask=None, positions=None, output_path=None, output_format='png'):
    """Plot the observed datas_ of an analysis, using the *Image* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    figsize
    plot_signal_to_noise_map
    ignore_config
    plot_psf
    plot_noise_map
    plot_image
    positions
    image : autolens.imaging.regular.Image
        Class containing the datas_, noise-mappers and PSF that are to be plotted.
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    plot_imaging_image = conf.instance.general.get('output', 'plot_imaging_image', bool)
    plot_imaging_noise_map = conf.instance.general.get('output', 'plot_imaging_noise_map', bool)
    plot_imaging_psf = conf.instance.general.get('output', 'plot_imaging_psf', bool)
    plot_imaging_signal_to_noise_map = conf.instance.general.get('output', 'plot_imaging_signal_to_noise_map', bool)

    if plot_imaging_image:
        plot_image(image=image, plot_origin=plot_origin, mask=mask, positions=positions, output_path=output_path,
                   output_format=output_format)

    if plot_imaging_noise_map:
        plot_noise_map(image=image,  plot_origin=plot_origin, mask=mask, output_path=output_path,
                       output_format=output_format)

    if plot_imaging_psf:
        plot_psf(image=image, plot_origin=plot_origin, output_path=output_path, output_format=output_format)

    if plot_imaging_signal_to_noise_map:
        plot_signal_to_noise_map(image=image, plot_origin=plot_origin, mask=mask, output_path=output_path,
                                 output_format=output_format)


def plot_image(image, plot_origin=True, mask=None, should_plot_border=False, positions=None,
               image_plane_pix_grid=None, as_subplot=False,
               units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
               cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
               cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
               title='Observed Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
               mask_pointsize=10, position_pointsize=30, grid_pointsize=1,
               output_path=None, output_format='show', output_filename='observed_image'):

    origin = get_origin(image=image, plot_origin=plot_origin)

    array_plotters.plot_array(array=image, origin=origin, mask=mask, should_plot_border=should_plot_border,
                              positions=positions, grid=image_plane_pix_grid, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_noise_map(image, plot_origin=True, mask=None, as_subplot=False,
                   units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                   cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                   cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                   title='Noise-Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                   mask_pointsize=10,
                   output_path=None, output_format='show', output_filename='noise_map_'):

    origin = get_origin(image=image.noise_map, plot_origin=plot_origin)

    array_plotters.plot_array(array=image.noise_map, origin=origin, mask=mask, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_psf(image, plot_origin=True, as_subplot=False,
             units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
             cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
             cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
             title='PSF', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
             output_path=None, output_format='show', output_filename='psf'):

    origin = get_origin(image=image.psf, plot_origin=plot_origin)

    array_plotters.plot_array(array=image.psf, origin=origin, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_signal_to_noise_map(image, plot_origin=True, mask=None, as_subplot=False,
                             units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                             cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                             cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                             title='Signal-To-Noise-Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                             mask_pointsize=10,
                             output_path=None, output_format='show', output_filename='signal_to_noise_map'):

    origin = get_origin(image=image, plot_origin=plot_origin)

    array_plotters.plot_array(array=image.signal_to_noise_map, origin=origin, mask=mask, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)


def get_origin(image, plot_origin):

    if plot_origin:
        return image.origin
    else:
        return None


from autolens.model.profiles import light_profiles as lp

import numpy as np

# noinspection PyUnusedLocal
class MockLightProfile(lp.LightProfile):

    def __init__(self, value, size=1):
        self.value = value
        self.size = size

    def intensities_from_grid(self, grid):
        return np.array(self.size*[self.value])


class MockMassProfile(object):

    def __init__(self, value):
        self.value = value

    def surface_density_from_grid(self, grid):
        return np.array([self.value])

    def potential_from_grid(self, grid):
        return np.array([self.value])

    def deflections_from_grid(self, grid):
        return np.array([self.value, self.value])


import numpy as np
import pytest

from autolens.data.imaging import image as im, convolution
from autolens.data.array.util import grid_util
from autolens.data.array import mask as msk
from autolens.lensing import lensing_image as li
from autolens.model.inversion import convolution as inversion_convolution


@pytest.fixture(name='image')
def make_image():
    psf = im.PSF(array=np.ones((3, 3)), pixel_scale=3.0, renormalize=False)
    return im.Image(np.ones((4, 4)), pixel_scale=3., psf=psf, noise_map=np.ones((4, 4)),
                    background_noise_map=2.0*np.ones((4,4)), poisson_noise_map=3.0*np.ones((4,4)),
                    exposure_time_map=6.0 * np.ones((4, 4)), background_sky_map=5.0 * np.ones((4, 4)))

@pytest.fixture(name="mask")
def make_mask():
    return msk.Mask(np.array([[True, True, True, True],
                              [True, False, False, True],
                              [True, False, False, True],
                              [True, True, True, True]]), pixel_scale=3.0)

@pytest.fixture(name="lensing_image")
def make_lensing_image(image, mask):
    return li.LensingImage(image=image, mask=mask)


class TestLensingImage(object):

    def test_attributes(self, image, lensing_image):
        assert image.pixel_scale == lensing_image.pixel_scale
        assert (image.psf == lensing_image.psf).all()
        assert (image.background_noise_map == lensing_image.background_noise_map).all()
        assert (image.poisson_noise_map == lensing_image.poisson_noise_map).all()
        assert (image.exposure_time_map == lensing_image.exposure_time_map).all()
        assert (image.background_sky_map == lensing_image.background_sky_map).all()

    def test__image_and_image_mapper(self, lensing_image):
        assert (lensing_image.image == np.ones((4, 4))).all()
        assert (lensing_image.noise_map == np.ones((4, 4))).all()
        assert (lensing_image.background_noise_map == 2.0*np.ones((4,4))).all()
        assert (lensing_image.poisson_noise_map == 3.0*np.ones((4,4))).all()
        assert (lensing_image.exposure_time_map == 6.0*np.ones((4,4))).all()
        assert (lensing_image.background_sky_map == 5.0*np.ones((4,4))).all()

    def test_masking(self, lensing_image):
        assert (lensing_image.noise_map_ == np.ones(4)).all()
        assert (lensing_image.background_noise_map_ == 2.0*np.ones(4)).all()
        assert (lensing_image.poisson_noise_map_ == 3.0*np.ones(4)).all()
        assert (lensing_image.exposure_time_map_ == 6.0*np.ones(4)).all()
        assert (lensing_image.background_sky_map_ == 5.0*np.ones(4)).all()

    def test_grids(self, lensing_image):
        assert lensing_image.grids.regular.shape == (4, 2)

        assert (lensing_image.grids.regular == np.array([[1.5, -1.5], [1.5, 1.5],
                                                       [-1.5, -1.5], [-1.5, 1.5]])).all()
        assert (lensing_image.grids.sub == np.array([[2.0, -2.0], [2.0, -1.0], [1.0, -2.0], [1.0, -1.0],
                                                     [2.0, 1.0], [2.0, 2.0], [1.0, 1.0], [1.0, 2.0],
                                                     [-1.0, -2.0], [-1.0, -1.0], [-2.0, -2.0], [-2.0, -1.0],
                                                     [-1.0, 1.0], [-1.0, 2.0], [-2.0, 1.0], [-2.0, 2.0]])).all()
        assert (lensing_image.grids.blurring == np.array([[4.5, -4.5], [4.5, -1.5], [4.5, 1.5], [4.5, 4.5],
                                                          [1.5, -4.5], [1.5, 4.5], [-1.5, -4.5], [-1.5, 4.5],
                                                          [-4.5, -4.5], [-4.5, -1.5], [-4.5, 1.5], [-4.5, 4.5]])).all()

    def test_padded_grids(self, lensing_image):

        padded_image_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(mask=np.full((6, 6), False),
                                                                                               pixel_scales=lensing_image.image.pixel_scales)

        assert (lensing_image.padded_grids.regular == padded_image_util).all()
        assert lensing_image.padded_grids.regular.image_shape == (4, 4)
        assert lensing_image.padded_grids.regular.padded_shape == (6, 6)

        padded_sub_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full((6, 6), False), pixel_scales=lensing_image.image.pixel_scales,
            sub_grid_size=lensing_image.grids.sub.sub_grid_size)

        assert lensing_image.padded_grids.sub == pytest.approx(padded_sub_util, 1e-4)
        assert lensing_image.padded_grids.sub.image_shape == (4, 4)
        assert lensing_image.padded_grids.sub.padded_shape == (6, 6)

        assert (lensing_image.padded_grids.blurring == np.array([[0.0, 0.0]])).all()

    def test_border(self, lensing_image):
        assert (lensing_image.border == np.array([0, 1, 2, 3])).all()

    def test_convolvers(self, lensing_image):
        assert type(lensing_image.convolver_image) == convolution.ConvolverImage
        assert type(lensing_image.convolver_mapping_matrix) == inversion_convolution.ConvolverMappingMatrix

    def test_subtract(self, lensing_image):
        subtracted_image = lensing_image - np.array([1, 0, 1, 0])
        assert isinstance(subtracted_image, li.LensingImage)
        assert (subtracted_image.psf == lensing_image.psf).all()
        assert subtracted_image.pixel_scale == lensing_image.pixel_scale

        assert subtracted_image == np.array([0, 1, 0, 1])

    def test__constructor_inputs(self):
        psf = im.PSF(np.ones((7, 7)), 1)
        image = im.Image(np.ones((51, 51)), pixel_scale=3., psf=psf, noise_map=np.ones((51, 51)))
        mask = msk.Mask.masked_for_shape_and_pixel_scale(shape=(51, 51), pixel_scale=1.0)
        mask[26, 26] = False

        lensing_image = li.LensingImage(image, mask, sub_grid_size=8, image_psf_shape=(5, 5),
                                        mapping_matrix_psf_shape=(3, 3), positions=[np.array([[1.0, 1.0]])])

        assert lensing_image.sub_grid_size == 8
        assert lensing_image.convolver_image.psf_shape == (5, 5)
        assert lensing_image.convolver_mapping_matrix.psf_shape == (3, 3)
        assert (lensing_image.positions[0] == np.array([[1.0, 1.0]])).all()


@pytest.fixture(name="lensing_hyper_image")
def make_lensing_hyper_image(image, mask):
    return li.LensingHyperImage(image=image, mask=mask, hyper_model_image=np.ones(4),
                                hyper_galaxy_images=[np.ones(4), np.ones(4)], hyper_minimum_values=[0.1, 0.2])


class TestLensingHyperImage(object):

    def test_attributes(self, image, lensing_hyper_image):

        assert image.pixel_scale == lensing_hyper_image.pixel_scale
        assert (image.psf == lensing_hyper_image.psf).all()
        assert (image.background_noise_map == lensing_hyper_image.background_noise_map).all()
        assert (lensing_hyper_image.hyper_model_image == np.ones(4)).all()
        assert (lensing_hyper_image.hyper_galaxy_images[0] == np.ones(4)).all()
        assert (lensing_hyper_image.hyper_galaxy_images[1] == np.ones(4)).all()
        assert lensing_hyper_image.hyper_minimum_values == [0.1, 0.2]

    def test__image_and_image_mapper(self, lensing_hyper_image):
        assert (lensing_hyper_image.image == np.ones((4, 4))).all()
        assert (lensing_hyper_image.noise_map == np.ones((4, 4))).all()
        assert (lensing_hyper_image.background_noise_map == 2.0*np.ones((4,4))).all()
        assert (lensing_hyper_image.poisson_noise_map == 3.0*np.ones((4,4))).all()
        assert (lensing_hyper_image.exposure_time_map == 6.0*np.ones((4,4))).all()
        assert (lensing_hyper_image.background_sky_map == 5.0*np.ones((4,4))).all()

    def test_masking(self, lensing_hyper_image):
        assert (lensing_hyper_image.noise_map_ == np.ones(4)).all()
        assert (lensing_hyper_image.background_noise_map_ == 2.0*np.ones(4)).all()
        assert (lensing_hyper_image.poisson_noise_map_ == 3.0*np.ones(4)).all()
        assert (lensing_hyper_image.exposure_time_map_ == 6.0*np.ones(4)).all()
        assert (lensing_hyper_image.background_sky_map_ == 5.0*np.ones(4)).all()